/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Class
    floatingBody

Description
    Single floating body with motion ODE

SourceFiles
    floatingBody.C

\*---------------------------------------------------------------------------*/

#ifndef floatingBody_H
#define floatingBody_H

#include "word.H"
#include "polyPatchID.H"
#include "fvMesh.H"
#include "sixDOFqODE.H"
#include "ODESolver.H"

#include "tetFemMatrices.H"
#include "tetPointFields.H"
#include "faceTetPolyPatch.H"

#include "OFstream.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// Class forward declarations
class polyMesh;

/*---------------------------------------------------------------------------*\
                       Class floatingBody Declaration
\*---------------------------------------------------------------------------*/

class floatingBody
{
    // Private data

        //- Name of floating body
        word name_;

        //- Reference to mesh
        const fvMesh& mesh_;


        // Motion variables

            //- Hull patch
            polyPatchID hull_;

            //- Barge slider patch: moving with barge translation and rotation
            polyPatchID bargeSlider_;

            //- Fixed slider patch: moving with barge translation
            polyPatchID fixedSlider_;

            //- 6-DOF motion equation
            sixDOFqODE sixDOF_;

            //- ODE solver
            autoPtr<ODESolver> solver_;

            //- ODE solver accuracy
            scalar epsilon_;

            //- Gravitational acceleration
            dimensionedVector g_;


        // Helper variables

            //- Current time index
            label curTimeIndex_;

            //- Force at the beginning of time-step
            vector oldForce_;

            //- Moment at the beginning of time-step
            vector oldMoment_;


        //- Output file for translation and rotation
        OFstream of_;


    // Private Member Functions

        //- Disallow default bitwise copy construct
        floatingBody(const floatingBody&);

        //- Disallow default bitwise assignment
        void operator=(const floatingBody&);


        //- Read gravity from environmental properties
        static dimensionedVector readGravity(const fvMesh& mesh);

        //- Check definition
        void check() const;

        //- Return force on the hull
        vector hullForce() const;

        //- Return force on the hull
        vector hullMoment() const;


public:


    // Constructors

        //- Construct from components
        floatingBody
        (
            const word& name,
            const fvMesh& mesh,
            const word& hullPatchName,
            const word& bargeSliderName,
            const word& fixedSliderName,
            const word& odeSolverName,
            const scalar epsilon
        );

        //- Construct from dictionary
        floatingBody
        (
            const word& name,
            const fvMesh& mesh,
            const dictionary& dict
        );


    // Destructor - default


    // Member Functions

        //- Return name
        const word& name() const
        {
            return name_;
        }

        //- Return ID of hull patch
        const polyPatchID& hull() const
        {
            return hull_;
        }

        //- Return ID of barge slider patch
        const polyPatchID& bargeSlider() const
        {
            return bargeSlider_;
        }

        //- Return ID of fixed slider patch
        const polyPatchID& fixedSlider() const
        {
            return fixedSlider_;
        }

        //- Set boundary motion
        void setMotion(tetPointVectorField& motionU);

        //- Write dictionary
        void writeDict(Ostream&) const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
