/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "threeDofMotion.H"
#include "addToRunTimeSelectionTable.H"
#include "motionSolver.H"
#include "volFields.H"

#include "tetFemMatrices.H"
#include "tetPointFields.H"
#include "faceTetPolyPatch.H"
#include "fixedValueTetPolyPatchFields.H"

#include "incompressible/transportModel/transportModel.H"
#include "incompressible/turbulenceModel/turbulenceModel.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
    defineTypeNameAndDebug(threeDofMotion, 0);

    addToRunTimeSelectionTable(dynamicFvMesh, threeDofMotion, IOobject);
} // End namespace Foam


// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

Foam::dimensionedVector Foam::threeDofMotion::readGravity(const fvMesh& mesh)
{
    Info << "\nReading environmentalProperties" << endl;

    IOdictionary environmentalProperties
    (
        IOobject
        (
            "environmentalProperties",
            mesh.time().constant(),
            mesh,
            IOobject::MUST_READ,
            IOobject::NO_WRITE
        )
    );

    dimensionedVector g(environmentalProperties.lookup("g"));

    return g;
}


Foam::vector Foam::threeDofMotion::hullForce() const
{
    const fvMesh& mesh = *this;

    const turbulenceModel& turbulence =
        mesh.lookupObject<turbulenceModel>("turbulenceProperties");

    volScalarField nuEff = turbulence.nuEff();

    const volScalarField& p = mesh.lookupObject<volScalarField>("p");
    const volVectorField& U = mesh.lookupObject<volVectorField>("U");

    // Get index of current patch
    const label curPatch = hullPatch_.index();

    vector pressureForce =
        gSum
        (
            p.boundaryField()[curPatch]*
            mesh.Sf().boundaryField()[curPatch]
        );

    vector viscousForce =
        gSum
        (
            -nuEff.boundaryField()[curPatch]*
            U.boundaryField()[curPatch].snGrad()*
            mesh.magSf().boundaryField()[curPatch]
        );

    vector totalForce = pressureForce + viscousForce;

    Info<< "Pressure force = " << pressureForce << nl
        << "Viscous force = " << viscousForce << nl
        << "Total force = " << totalForce << endl;

    return totalForce;
}


Foam::vector Foam::threeDofMotion::hullMoment() const
{
    const fvMesh& mesh = *this;

    const turbulenceModel& turbulence =
        mesh.lookupObject<turbulenceModel>("turbulenceProperties");

    volScalarField nuEff = turbulence.nuEff();

    const volScalarField& p = mesh.lookupObject<volScalarField>("p");
    const volVectorField& U = mesh.lookupObject<volVectorField>("U");

    // Get index of current patch
    const label curPatch = hullPatch_.index();

    vector pressureMoment =
        gSum
        (
            (mesh.Cf().boundaryField()[curPatch] - equation_.X().value())
          ^ (
                p.boundaryField()[curPatch]*
                mesh.Sf().boundaryField()[curPatch]
            )
        );

    vector viscousMoment =
        gSum
        (
            (mesh.Cf().boundaryField()[curPatch] - equation_.X().value())
          ^ (
                -nuEff.boundaryField()[curPatch]*
                U.boundaryField()[curPatch].snGrad()*
                mesh.magSf().boundaryField()[curPatch]
            )
        );

    vector totalMoment = pressureMoment + viscousMoment;

    Info<< "Pressure moment = " << pressureMoment << nl
        << "Viscous moment = " << viscousMoment << nl
        << "Total moment = " << totalMoment << endl;

    return totalMoment;
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::threeDofMotion::threeDofMotion(const IOobject& io)
:
    dynamicFvMesh(io),
    dict_
    (
        IOdictionary
        (
            IOobject
            (
                "dynamicMeshDict",
                time().constant(),
                *this,
                IOobject::MUST_READ,
                IOobject::NO_WRITE
            )
        ).subDict(typeName + "Coeffs")
    ),
    equation_
    (
        IOobject
        (
            dict_.lookup("name"),
            time().timeName(),
            *this,
            IOobject::MUST_READ,
            IOobject::AUTO_WRITE
        )
    ),
    solver_
    (
        ODESolver::New
        (
            dict_.lookup("solver"),
            equation_
        )
    ),
    epsilon_(readScalar(dict_.lookup("epsilon"))),
    g_(readGravity(*this)),
    motionPtr_(motionSolver::New(*this)),
    hullPatch_(word(dict_.lookup("hullPatchName")), boundaryMesh()),
    curTimeIndex_(-1),
    oldForce_(vector::zero)
{
    // Check that the hull patch has been found
    if (!hullPatch_.active())
    {
        FatalErrorIn("threeDofMotion::threeDofMotion(const IOobject& io)")
            << "Cannot find hull patch "<< hullPatch_.name() << nl
            << "Valid patches are: " << boundaryMesh().names()
            << abort(FatalError);
    }
}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::threeDofMotion::~threeDofMotion()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

bool Foam::threeDofMotion::update()
{
    // Grab force
    vector curForce = hullForce();

    equation_.force().value() = 0.5*(curForce + oldForce_)
        + equation_.mass().value()*g_.value();

    if (curTimeIndex_ == -1 || curTimeIndex_ < time().timeIndex())
    {
        Info << "Storing current force as old force" << endl;
        oldForce_ = curForce;
        curTimeIndex_ = time().timeIndex();
    }

    solver_->solve
    (
        time().value(),
        time().value() + time().deltaT().value(),
        epsilon_,
        time().deltaT().value()
    );

    // Set the motion onto the motion patch
    tetPointVectorField& motionU =
        const_cast<tetPointVectorField&>
        (
            this->objectRegistry::lookupObject<tetPointVectorField>("motionU")
        );

    fixedValueTetPolyPatchVectorField& motionUHullPatch =
        refCast<fixedValueTetPolyPatchVectorField>
        (
            motionU.boundaryField()[hullPatch_.index()]
        );

    Info << "Boundary velocity = " << equation_.U().value() << endl;

    motionUHullPatch == equation_.U().value();

    fvMesh::movePoints(motionPtr_->newPoints());

    // Mesh motion only - return false
    return false;
}


// ************************************************************************* //
